#!/usr/bin/env python3
# vi: set softtabstop=2 ts=2 sw=2 expandtab:
# pylint:

import os
import os.path
import sys
import re
import configparser
import argparse
import json
import urllib.parse
from datetime import datetime, timezone
import requests
import jinja2
from enc_hosts.version import version

# ---------------------------------------------------------------------------
#                                                               CONSTANTS
# ---------------------------------------------------------------------------

# default pagination
# TODO: check if Foreman > 1.24 honours "all"
ALL = 99999

# ---------------------------------------------------------------------------
#                                                               FUNCTIONS
# ---------------------------------------------------------------------------

def fatal(msg, rc=1):
  print(msg, file=sys.stderr)
  sys.exit(rc)

def error(msg):
  print(msg, file=sys.stderr)

# pylint: disable=unused-argument
def print_response(r, *args, **kwargs):
  print("Receiving from: {}".format(r.url), file=sys.stderr)

def paginated_query(url, auth, query=None, options=None):

  # set defaults and parse options if any
  opts = {
    'tls_verify': True,
    'per_page': ALL
  }
  if options:
    opts.update(options)

  # only print feedback if paginating
  hooks = None if opts['per_page'] == ALL else {'response':print_response}

  # set up base query URL
  if query == "":
    qurl = f"{url}?"
  else:
    qurl = f"{url}?{query}&"

  # iterate
  done = False
  page = 1
  while not done:
    try:
      response = requests.get(
        "{}page={}&per_page={}".format(qurl, page, opts['per_page']),
        verify=opts['tls_verify'], auth=auth, hooks=hooks
      )
    except Exception as e:
      print(e.traceback.format_exc())
    r = response.json()

    # continue while results dict not empty
    if 'results' in r and r['results']:
      yield r['results']
      page += 1
    else:
      done = True
      break

def cache_valid(path):
  if os.path.exists(path):
    return True
  return False

def read_from_cache(path):
  with open(path, 'r') as file:
    return json.loads(file.read())

def write_to_cache(path, hosts):
  with open(path, 'w') as file:
    file.write(json.dumps(hosts))

def query_facts_from_hostid(auth, url, host_fqdn, options=None):

  facts_url = url + "/" + str(host_fqdn) + "/facts"
  facts = {}
  for results in paginated_query(facts_url, auth, options=options):

    # update facts with results
    facts.update(results[host_fqdn])

  return facts

def query_from_enc(auth, url, querystr, return_facts, options=None):

  # get list of hosts
  hosts = {}
  for results in paginated_query(url, auth, query=querystr, options=options):

    # add results to hosts dictionary
    ## if return_facts is true, then do a second query for facts
    for host in results:
      hosts[host['name']] = host
      host['x_params_map'] = {}
      for parameter in host['all_parameters']:
        host['x_params_map'][parameter['name']] = parameter['value']
      # if they've asked for the facts, run a second API call for this host
      if return_facts:
        facts = query_facts_from_hostid(
          auth, url, host['name'], options=options
        )
        host['facts'] = facts

  return hosts

# ---------------------------------------------------------------------------
#                                                               JINJA FILTERS
# ---------------------------------------------------------------------------

def regex(str, regex):
  regexc = re.compile(regex)
  return regexc.match(str)

# output iso8601 timestamp for either UTC or local time
def timestamp(zone):
  if zone == 'utc':
    return datetime.now().astimezone(timezone.utc).replace(microsecond=0).isoformat()
  # default to local timezone if not UTC
  return datetime.now().astimezone().replace(microsecond=0).isoformat()

# ---------------------------------------------------------------------------
#                                                               MAIN
# ---------------------------------------------------------------------------

def main():

  # program name
  _tag = "enc-hosts"

  # program description
  description = '''
  Pull host information from ENC (enterprise node classifier) and run through
  given template.
  '''
  epilog = "Version {}.  For more information please visit project homepage: " \
    "https://git.computecanada.ca/dleske/enc-hosts".format(version)

  # get command-line options
  # Usage: $0 [-c config]
  parser = argparse.ArgumentParser(prog=_tag, description=description, epilog=epilog)
  parser.add_argument("-c", "--config", type=str,
                      help="Configuration file",
                      default=_tag + ".conf")
  parser.add_argument("--use-cache",
                      help="Use cached results if available",
                      action="store_true")
  pag_args = parser.add_mutually_exclusive_group()
  pag_args.add_argument("--pagination", type=int, default=20, metavar='RESULTS',
                      help="Paginate results with RESULTS per page")
  pag_args.add_argument('--no-pagination', action='store_true', help='Do not paginate')
  parser.add_argument("--search", help="Host search string")
  parser.add_argument("--suffix", type=str,
                      help="Suffix to apply to rendered templates",
                      default=".out")
  parser.add_argument("--facts",
                      help="Also pull facts for each host (CAUTION: very resource intensive)",
                      action="store_true")
  parser.add_argument("templates", metavar='TEMPLATE',
                      help="Template(s) to render", nargs="+")
  args = parser.parse_args()

  # determine pagination
  pagination = ALL if args.no_pagination else args.pagination

  # initialize configuration
  config = configparser.ConfigParser(delimiters='=')
  configs_read = config.read(args.config)
  if len(configs_read) == 0:
    fatal("Unable to read configuration file: {}".format(args.config))

  # set up to skip TLS verification if so configured
  tls_verify = True
  if config['main']['skip_verification']:
    # pylint: disable=import-outside-toplevel
    import urllib3
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    tls_verify = False

  # build query
  base_url = config['main']['server_url'] + '/api/hosts'
  query = 'include=all_parameters'
  auth = (config['main']['username'], config['main']['password'])
  if args.search:
    encoded_search = urllib.parse.quote_plus(args.search)
    query += '&search=' + encoded_search
    cachefile = encoded_search
  else:
    cachefile = 'hosts'

  # determine cachepath
  cachepath = None
  if 'cache_dir' in config['main']:
    cachedir = os.path.expanduser(config['main']['cache_dir'])
    if os.path.exists(cachedir):
      cachepath = '{}/{}'.format(cachedir, cachefile)

  # check templates exist
  templatedir = config['main'].get('template_dir', 'templates')
  templates = []
  for template in args.templates:
    basepath = f"{templatedir}/{template}"
    rendered = f"{template}{args.suffix}"
    if os.path.isfile(f"{basepath}.jinja"):
      templates.append((f"{template}.jinja", rendered))
    elif os.path.isfile(f"{basepath}.j2"):
      templates.append((f"{template}.j2", rendered))
    else:
      fatal(f"Found neither {basepath}.jinja nor {basepath}.j2")

  # if using cache and cache is valid
  if args.use_cache and cache_valid(cachepath):
    print("Reading from cache", file=sys.stderr)
    hosts = read_from_cache(cachepath)
  else:
    hosts = query_from_enc(
      auth, base_url, query, args.facts,
      options={'tls_verify': tls_verify, 'per_page': pagination}
    )

    # check that we have something
    if not hosts:
      fatal("Query returned no results")

    # if caching
    if cachepath:
      write_to_cache(cachepath, hosts)

  # template loading
  templateloader = jinja2.FileSystemLoader(templatedir)

  # templating environment
  jenv = jinja2.Environment(loader=templateloader)

  # add filters
  jenv.filters['timestamp'] = timestamp

  # add custom tests
  jenv.tests['regex'] = regex

  # loop through given templates
  for (template_fn, rendered_fn) in templates:

    print("Rendering template {} to {}".format(template_fn, rendered_fn), file=sys.stderr)

    try:
      jtmpl = jenv.get_template(template_fn)
    except jinja2.exceptions.TemplateNotFound:
      error("Template not found: {template_fn} in template directory {templatedir}")
    else:
      with open(rendered_fn, 'w') as rendered:
        rendered.write(jtmpl.render(hosts=hosts))

# invoke main function
if __name__ == '__main__':
  sys.exit(main())
